"""
One of the main purposes of using pipeline managers like Luigi, Dask, Airflow is
to store the result of heavy computation tasks for further reuse. It happens
often that the process of developing the whole pipeline is very dynamic. The
code of a task can be explicitly changed or the task behavior can change due
to external changes, like an update of a python package. It turns out that
one cannot be certain that the stored result of the task is up-to-date with
the code. This may invalidate any test/validation process.
The solution we propose here is to generate a hash of the code of each task.
In the case of using Luigi, we assume that if the target was generated with the
same task parameters and it has the same hash code we can safely use this
target (of course, excluding hash collision hypothesis). We observe that some
pipeline managers also keep track of the data/inputs hash. This module can
cooperate with these managers as well.
Finally, the practical objective of this module is to obtain a hash code of a
given function statically, i.e., without computing the python code itself.
Due to its statical nature, some case cannot be supported, like import and
functions definitions inside if statements.
The method get_bytecode_tree performs all the hard work and return nested lists
of bytecodes that may be useful for inspection.
The method get_function_hash uses get_bytecode_tree to return a hash for a
given function.
"""

import dis
import inspect

from pycarol.pipeline.utils import (
    int_to_bytes,
    flat_list,
    is_builtin,
    is_function,
    is_code,
    enumerate_with_context
)

#TODO: move this module to a dedicated repository
DEBUG_MODE = False

from pycarol.pipeline.utils.hash_versioning.inspect_bytecode import (
    get_name_and_code_of_MAKE_FUNCTION,
    get_name_and_object_of_IMPORT_NAME,
    get_name_and_object_of_CALL_FUNCTION,
    get_name_and_object_of_LOAD_ATTR,
)


def _get_consts_hash(f) -> bytes:
    """Returns hash code of local consts"""
    if hasattr(f, '__code__'):  # is function object
        consts_list = list(f.__code__.co_consts)
    elif hasattr(f, 'co_code'):  # is code object
        consts_list = list(f.co_consts)
    else:
        raise TypeError

    for i, v in enumerate(consts_list):
        if isinstance(v, str) and "<locals>" in v:
            consts_list[i] = v.split('.')[-1]
    consts_tuple = tuple(consts_list)
    return int_to_bytes(hash(consts_tuple))

def process_op(
        context,
        defined_functions,
        code_set,
        called_functions,
        parent_function,
        robust
        ):
    """
    Process instruction updating defined_functions, code_set and
    called_functions accordingly
    Args:
        context: ix, inst, instructions. bytecode itself
        defined_functions: set of functions defined until now
        code_set: set of functions already inspected for hash versioning purpose
        called_functions: set of called functions until now in this bytecode
        parent_function: function object of this bytecode
        robust: set true to ignore internal assertion errors

    Returns:

    """
    ix, inst, instructions = context
    if inst.opname == "MAKE_FUNCTION":
        defined_functions.update(
            get_name_and_code_of_MAKE_FUNCTION(*context))

    if inst.opname == "IMPORT_NAME": #TODO: some tests are failing for IMPORT_NAME
        defined_functions.update(
            get_name_and_object_of_IMPORT_NAME(*context))

    if "CALL_FUNCTION" in inst.opname:
        try:
            _, son_function = get_name_and_object_of_CALL_FUNCTION(
                ix,inst,instructions,
                parent_function,
                defined_functions
            )
        except AssertionError as error:
            if robust:
                # print(error)
                return  # do not update code_set and called_functions
            else:
                raise(error)

        if son_function not in code_set:
            code_set.add(son_function)
            called_functions.add(son_function)
    return

def traverse_code(parent_function: 'function', code_set,
                  ignore_not_implemented=True) -> list:
    """
    This method recursively traverse the bytecodes of top_function entering
    in every function call found. At every level it returns a list of
    bytecodes of respective functions called.

    Args:
        parent_function: function to be analyzed at this recursion level
        code_set: set of functions already analyzed
        ignore_not_implemented: if True, will generate bytecode_tree where it is
        possible. Where it is not possible, will return empty list.

    Returns:
        bytecode_tree: nested lists of bytecode

    """

    if is_builtin(parent_function):
        # dis cannot get instructions for builtins
        # return function name instead of bytecode
        return [parent_function.__name__]

    called_functions = set()
    defined_functions: dict = {}

    # linear search on bytecodes for called functions
    instructions = list(dis.get_instructions(parent_function))
    for context in enumerate_with_context(instructions):
        try:
            process_op(context,defined_functions,code_set,called_functions,
                       parent_function,ignore_not_implemented)
        except NotImplementedError as e:
            if ignore_not_implemented:
                continue
            else:
                raise e

    # recursion step
    code_list = [traverse_code(f, code_set, ignore_not_implemented) for f in
                 called_functions]

    # leaf node step
    if is_function(parent_function):

        function_code: list = b''.join([
            parent_function.__code__.co_code,
            _get_consts_hash(parent_function),
            int_to_bytes(hash(
                dict(inspect.getmembers(parent_function)).get('__defaults__',0)
            )),
        ])
    elif is_code(parent_function):  # parent_function is code object
        function_code: list = b''.join([
            parent_function.co_code,
            _get_consts_hash(parent_function),
            # missing default parameter information
            # asbytes(hash(
            #     dict(inspect.getmembers(parent_function))['__defaults__']
            # )),
        ])
    else:
        raise TypeError("parent_function should be either function or code.")

    code_list.append(function_code)
    return code_list


def get_bytecode_tree(
        top_function: 'function',
        ignore_not_implemented=False,
) -> list:
    """
    Args:
        top_function: analyzed function
        ignore_not_implemented: if True, will generate bytecode_tree where it is
        possible. Where it is not possible, will return empty list.

    Returns:
        bytecode_tree: nested lists of bytecode

    """
    #TODO: nested, builtin, call_kwargs are failing interprocesses test
    from functools import partial
    if isinstance(top_function,partial):
        top_function = top_function.func

    assert is_function(top_function), 'argument should be a function'
    code_set = set()  # "global" memoization set
    bytecode_tree = traverse_code(top_function, code_set,ignore_not_implemented)

    return bytecode_tree

def get_function_hash(f: 'function', ignore_not_implemented=False) -> int:
    """
    Module main function. It returns a proper hash for the given function.
    Args:
        f: function to be hashed
        ignore_not_implemented: if True, will consider bytecode_tree where it is
        possible. Where it is not possible, will replace by an empty list.
        and compute a hash anyway. Some part of the code may be ignored.

    Returns:
        h: a hash number for the given function

    """
    bytecode_nested_list = get_bytecode_tree(f, ignore_not_implemented)
    bytecode_flat_list = flat_list(bytecode_nested_list)
    bytecode_flat_list = [
        bytecode.encode()
        if isinstance(bytecode,str)
        else bytecode
        for bytecode in bytecode_flat_list
    ]
    
    from hashlib import sha256
    hashable_bytecode = b''.join(bytecode_flat_list)
    h = sha256(hashable_bytecode).hexdigest()
    return h





from ..utils import get_reverse_dag, breadth_first_search


def nodes_layout(
        dag:dict,
        align_on_leafs=True,
        mode='force-directed',
        **layout_params
) -> dict:
    """
    Returns layout with x,y coordinates to plot DAG. Available modes are:
      'levels-directed': x positions correspond to the level of the node in DAG.
      'force-directed': computed using fruchterman_reingold_layout method of
      networkx library.
      'kamada-kawai': computed using networkx library
    If align on_leafs is set to true, DAG is first reversed, so that the
    output of the pipeline is on the right and inputs on the left in
    'levels-directed' mode.
    Args:
        dag: dict encoding a DAG structure
        align_on_leafs: layout boolean parameter
        mode: string defining layout mode

    Returns:
        layout: a dict whose keys are DAG nodes an values are (x,y) of each
        node.

    """
    layout_x = {}
    if align_on_leafs:
        dag = get_reverse_dag(dag)
    for i, nodes in enumerate(breadth_first_search(dag)):
        for j, node in enumerate(nodes):
            # overwrite previous levels and keep only last one
            layout_x[node] = i

    levels = sorted(v for v in layout_x.values())
    layout = {}
    for l in levels:
        y = 0
        for node, x in layout_x.items():
            if x == l:
                layout[node] = (x, y)
                y += 1

    if mode == "force-directed":
        import networkx as nx
        graph = nx.DiGraph()
        graph_edges = [(k, vi) for k, v in dag.items() for vi in v]
        graph.add_edges_from(graph_edges)
        default_layout = layout.copy()
        from pycarol.pipeline.utils import find_root_in_dag, find_leaf_in_dag
        root_nodes = find_root_in_dag(dag)
        leaf_nodes = find_leaf_in_dag(dag)
        fixed_nodes = root_nodes + leaf_nodes
        layout = nx.fruchterman_reingold_layout(
            graph,
            pos=default_layout,  # we use levels layout as default position
            fixed=fixed_nodes,
            **layout_params
        )
    elif mode == "kamada-kawai":
        import networkx as nx
        graph = nx.DiGraph()
        graph_edges = [(k, vi) for k, v in dag.items() for vi in v]
        graph.add_edges_from(graph_edges)
        layout = nx.kamada_kawai_layout(
            graph,
            **layout_params
        )

    return layout


def edges_layout(dag:dict, layout:dict) -> list:
    """
    Given a dag network and the positions of each one of its nodes,
    this function creates the positions of the edges.
    Args:
        dag: dict encoding a DAG
        layout: dict containing nodes x,y positions

    Returns:
        edges: list of positions, each one in format ( (x0,y0), (x1,y1) )
    """

    edges = []
    for source_node in layout:
        for target_node in dag[source_node]:
            edges.append((layout[source_node],layout[target_node]),)
    return edges

def get_task_id(t):
    return t.task_id

def get_task_family(t):
    # returns task namespace
        if '.' in t.task_id:
            return t.task_id.split('.')[-2]
        else:
            return ""

def get_task_name(t):
    return t.task_id.split('.')[-1].split('_')[0] # name of the task class

def get_complete(t):
    # return t.complete()
    # make it dummy, because efficient get_complete needs pipeline object
    return False


def get_target_hash_version(t):
    try:
        metadata = t.load_metadata()
    except:
        return ""
    return metadata.get('hash_version',"")

def get_task_hash_version(t):
    try:
        h = t.hash_version()
    except:
        h = ""
    return h
def get_task_version(t):
    try:
        h = t.version
    except:
        h = ""
    return h

def get_target_version(t):
    try:
        metadata = t.load_metadata()
    except:
        return ""
    return metadata.get('version',"")

#TODO: implement dash viewer
def make_nodes_data_source(nodes_layout) -> dict:
    """
    Creates a bokeh compatible data source encoding nodes plotting
    properties. Returns this data source in bokeh compatible dict format
    Args:
        nodes_layout: dict containing nodes as keys and its positions as
        values. In this function nodes are luigi tasks.

    Returns:
        data_source: bokeh compatible dict containing the columns: x, y,
        task_id, task_family, task_name, complete, tasklog, hash_version

    """

    data = dict(
        x=[],
        y=[],
        task_id=[],
        task_family=[],
        task_name=[],
        complete=[],
        target_version=[],
        task_version=[],
    )
    for k,(x,y) in nodes_layout.items():
        data['x'].append(x)
        data['y'].append(y)
        data['task_id'].append(get_task_id(k))
        data['task_family'].append(get_task_family(k))
        data['task_name'].append(get_task_name(k))
        data['complete'].append(get_complete(k))
        data['target_version'].append(get_target_version(k))
        data['task_version'].append(get_task_version(k))

    
    v_gen = (v for v in data.values())
    first_v = next(v_gen)
    for v in v_gen:
        if len(first_v) != len(v):
            raise ValueError("all items in data should have the same length")

    return data

def make_edges_data_source(edges_layout) -> dict:
    """
    Creates a bokeh segment glyph compatible data source encoding edges plotting
    properties. Returns this data source in bokeh compatible dict format
    Args:
        edges_layout: list containing edges coordinates.

    Returns:
        data_source: bokeh segment glyph compatible dict containing the
        columns: x0, y0, x1, y1

    """

    data = dict(
        x0=[],
        y0=[],
        x1=[],
        y1=[]
    )
    for ((x0,y0,),(x1,y1)) in edges_layout:
        data['x0'].append(x0)
        data['y0'].append(y0)
        data['x1'].append(x1)
        data['y1'].append(y1)

    return data




from pycarol.pipeline import Task
from pycarol.pipeline.utils import build_dag
from pycarol.pipeline.utils import breadth_first_search, get_reverse_dag, find_root_in_dag
import warnings

import luigi

def luigi_get_sons(task) -> list:
    """
    Returns a list of required tasks. This is used in build_dag
    Args:
        task: luigi Task

    Returns:
        l: list of luigi Task

    """
    if isinstance(task,tuple):
        # this case happens when we use local params in tasks require
        task,params = task
    return task.requires_list


def get_dag_from_task(task:list) -> dict:
    """
    Wrapper around generic build_dag.
    Args:
        task: list of proper luigi tasks

    Returns:
        dag: dict encoding a DAG

    """
    dag = build_dag(task,luigi_get_sons)
    return dag

def get_instances_from_classes(dag:dict,params:dict):
    """Returns a dag of task instances, given a dag of task classes and pipeline params."""
    instances_dag = {}
    for k,v in dag.items():
        task_params = params
        if isinstance(k,tuple):
            k,local_params = k
            task_params.update(local_params)
        task_list=[]
        for t in v:
            task_params = params
            if isinstance(t,tuple):
                t,local_params = t
                task_params.update(local_params)
            task_list.append(t(**task_params))
        instances_dag[k(**params)] = task_list
    return instances_dag

def downstream_complete(dag,top_nodes,downstream_complete_dict):
    """Recursively traverses dag starting from top_nodes to update downstream_complet_dict"""
    #TODO: reimplement using breadth_first_search
    for task in top_nodes:
        if task in downstream_complete_dict:
            continue
        sons = dag[task]
        if sons: # recursion step
            downstream_complete_dict[task] = task.complete() and \
                all([
                    downstream_complete(dag,[t],downstream_complete_dict) 
                    for t in sons
                    ])
        else: # stop recursion step
            downstream_complete_dict[task] = task.complete()

def _tasks_are_class(tasks):
    for t in tasks:
        if not issubclass(t,Task):
            return False
    else:
        return True

def _tasks_are_instance(tasks):
    for t in tasks:
        if not isinstance(t,Task):
            return False
    else:
        return True



class Pipe(object):
    """
    This class should be used to compose a pipeline given a list of tasks. It
    contains many methods to interact with the pipeline as a whole.
    """
    def __init__(self, tasks: list, params:dict):
        assert isinstance(tasks,list)
        assert params is None or isinstance(params,dict)
        assert _tasks_are_instance(tasks) or _tasks_are_class(tasks)

        self.params = params
        self.top_nodes = tasks # top nodes are root nodes
        self.dag = get_dag_from_task(tasks)
        if _tasks_are_instance(tasks):
            raise NotImplementedError(f"Currently, we cannot create Pipe objects from instantiated tasks.")
        if _tasks_are_instance(tasks) and params is not None:
            warnings.warn("params will not be used because tasks are already initialized")
        if _tasks_are_class(tasks):
            self.top_nodes = [t(**self.params) for t in self.top_nodes]
            self.dag = get_instances_from_classes(self.dag,self.params)


        self.rev_dag = get_reverse_dag(self.dag)
        self.leaf_nodes = find_root_in_dag(self.rev_dag) #  leaf nodes are root nodes of rev dag
        self.all_tasks = [k for k in self.dag]
        

    def remove_all(self):
        """Remove all targets related to this pipeline."""
        for t in self.all_tasks:
            try:
                t.remove()
            except FileNotFoundError:
                pass

    def remove_upstream(self, tasks:list):
        """Remove all targets in this pipeline that depend on the given tasks."""
        assert isinstance(tasks,list)
        traverse_dag_generator = breadth_first_search( self.rev_dag,tasks)
        for task_list in traverse_dag_generator:
            for t in task_list:
                try:
                    t.remove()
                except FileNotFoundError:
                    pass
    
    def remove_orphans(self):
        """Remove all targets for which respective downstream targets are not complete"""

        downstream_complete_dict = {}
        downstream_complete(self.dag,self.top_nodes,downstream_complete_dict)

        for t, is_downstream_complete in downstream_complete_dict.items():
            if  not is_downstream_complete:
                try:
                    t.remove()
                except FileNotFoundError:
                    pass

    def remove_obsolete(self):
        """Remove all targets whose hash_versions do not match to current version"""
        return

    def update_all_complete_status(self):
        """ Updates a dictionary whose keys are task objects and values are
        True if target was found."""
        from pycarol import Carol, Storage
        st = Storage(Carol())
        target_files = st.files_storage_list(prefix='pipeline/')
        self.all_complete_status = {t: t.output().path in target_files for t in
                self.all_tasks}

    def get_task_complete(self,task):
        """ Returns True if task is complete accordingly to
        self.all_complete_status dictionary"""
        assert task in self.all_tasks, f"Task {task} not found in this pipeline"
        return self.all_complete_status[task]

    def run(self):
        """Run the whole pipeline"""
        tasks = [t for t in self.top_nodes]
        luigi.build(tasks,local_scheduler = True)

    def get_dag(self):
        return self.dag

    def get_task_by_id(self,task_id):
        for t in self.all_tasks:
            if task_id == t.task_id:
                return t
        else:
            raise KeyError(f"{task_id} not found in this pipeline.")

    def get_matching_tasks(self,task):
        matching_tasks = [isinstance(t, task) for t in self.all_tasks]
        matching_tasks = [t for i, t in enumerate(self.all_tasks) if
                          matching_tasks[i]]
        return matching_tasks


    def assert_task_is_unique(self,task):
        matching_tasks = self.get_matching_tasks(task)
        assert len(matching_tasks) > 0
        if len(matching_tasks) == 1:
            return None
        params = matching_tasks[0].get_execution_params()
        for t in matching_tasks[1:]:
            params_i = t.get_execution_params()
            assert params_i == params
        return None


    def get_task_instance(self,task):
        self.assert_task_is_unique(task)
        matching_tasks = self.get_matching_tasks(task)
        return matching_tasks[0]





from .ApiKeyAuth_cloner import ApiKeyAuthCloner


class ApiKeyAuth:
    def __init__(self, api_key):
        self.api_key = api_key
        self.carol = None
        self.connector_id = None
        
    def set_connector_id(self, connector_id):
        self.connector_id = connector_id
    
    def cloner(self):
        return ApiKeyAuthCloner(self)

    def login(self, carol):
        self.carol = carol
        pass

    def authenticate_request(self, headers):
        headers['x-auth-key'] = self.api_key
        headers['x-auth-connectorid'] = self.connector_id


""" PyCarol - Connecting Carol to Python

"""

import os
import tempfile

__version__ = '2.23.0'

__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), 'carolina/cache')

__CONNECTOR_PYCAROL__ = 'f9953f6645f449baaccd16ab462f9b64'

_CAROL_METADATA = ['mdmCounterForEntity', 'mdmId', 'mdmLastUpdated',
                   'mdmConnectorId', 'mdmDeleted']


from .carol import Carol
from .staging import Staging
from .connectors import Connectors
from .query import Query
from .storage import Storage
from .carolina import Carolina
from .tasks import Tasks
from .data_models import DataModel, DataModelView
from .logger import CarolHandler
from .auth.PwdAuth import PwdAuth
from .auth.ApiKeyAuth import ApiKeyAuth




import os
import pickle
import calendar
import gzip
import pandas as pd
import botocore
from .. import __TEMP_STORAGE__
from ..utils.miscellaneous import prettify_path, _attach_path, _FILE_MARKER
from collections import defaultdict

class StorageAWSS3:
    def __init__(self, carol, carolina):
        self.carol = carol
        self.client = None
        self.bucket = None
        self.carolina = carolina

    def _init_if_needed(self):
        if self.client is not None:
            return

        self.client = self.carolina.get_client()
        self.bucket = self.client.Bucket(self.carolina.cds_app_storage_path['bucket'])
        if not os.path.exists(__TEMP_STORAGE__):
            os.makedirs(__TEMP_STORAGE__)

    def save(self, name, obj, format='pickle', parquet=False, cache=True):
        self._init_if_needed()
        s3_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
        local_file_name = os.path.join(__TEMP_STORAGE__, s3_file_name.replace("/", "-"))

        if parquet:
            if not isinstance(obj, pd.DataFrame):
                raise ValueError("Object to be saved as parquet must be a DataFrame")
            obj.to_parquet(local_file_name)
        elif format == 'joblib':
            import joblib

            if not cache:
                from io import BytesIO
                with BytesIO() as buffer:
                    joblib.dump(obj, buffer)
                    buffer.seek(0)
                    self.bucket.upload_fileobj(buffer, s3_file_name)
                return
            else:
                joblib.dump(obj, local_file_name)
        elif format == 'pickle':
            with gzip.open(local_file_name, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        elif format == 'file':
            local_file_name = obj
        else:
            raise ValueError("Supported formats are pickle, joblib or file")

        self.bucket.upload_file(local_file_name, s3_file_name)
        os.utime(local_file_name, None)

    def load(self, name, format='pickle', parquet=False, cache=True, storage_space='app_storage', columns=None):
        self._init_if_needed()
        if storage_space == 'app_storage':
            s3_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
        else:
            s3_file_name = name
        local_file_name = os.path.join(__TEMP_STORAGE__,s3_file_name.replace("/", "-"))

        obj = self.bucket.Object(s3_file_name)
        if obj is None:
            return None

        if not cache and format == 'raw':
            from io import BytesIO
            buffer = BytesIO()
            self.bucket.download_fileobj(s3_file_name, buffer)
            return buffer

        elif not cache and format == 'joblib':
            import joblib
            from io import BytesIO
            with BytesIO() as data:
                self.bucket.download_fileobj(s3_file_name, data)
                data.seek(0)
                return joblib.load(data)

        has_cache = cache and os.path.isfile(local_file_name)

        if has_cache:
            localts = os.stat(local_file_name).st_mtime
        else:
            localts = 0

        try:
            s3ts = calendar.timegm(obj.last_modified.timetuple())
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                return None
            raise e

        # Local cache is outdated
        if localts < s3ts:
            self.bucket.download_file(s3_file_name, local_file_name)

        if os.path.isfile(local_file_name):
            if parquet:
                return pd.read_parquet(local_file_name,columns=columns)
            elif format == 'joblib':
                import joblib
                return joblib.load(local_file_name)
            elif format == 'pickle':
                with gzip.open(local_file_name, 'rb') as f:
                    return pickle.load(f)
            elif format == 'file':
                return local_file_name
            else:
                raise ValueError("Supported formats are pickle, joblib or file")
        else:
            return None

    def exists(self, name):
        self._init_if_needed()
        s3_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"

        obj = self.bucket.Object(s3_file_name)
        if obj is None:
            return False

        try:
            calendar.timegm(obj.last_modified.timetuple())
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                return False

        return True

    def delete(self, name):
        self._init_if_needed()
        s3_file_name = f"{self.carolina.cds_app_storage_path['path']}/{name}"
        obj = self.bucket.Object(s3_file_name)
        if obj is not None:
            obj.delete()

        local_file_name = os.path.join(__TEMP_STORAGE__,s3_file_name.replace("/", "-"))
        if os.path.isfile(local_file_name):
            os.remove(local_file_name)

    def build_url_parquet_golden(self, dm_name):
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            dm_name=dm_name,
            tenant_id=tenant_id,
            bucket=self.carolina.cds_app_storage_path["bucket"]
        )
        return f's3://{self.carolina.cds_app_storage_path["bucket"]}/{self.carolina.cds_golden_path["path"].format(**template)}'

    def build_url_parquet_view(self, view_name):
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            view_name=view_name,
            tenant_id=tenant_id,
            bucket=self.carolina.cds_app_storage_path["bucket"]
        )
        return f's3://{self.carolina.cds_app_storage_path["bucket"]}/{self.carolina.cds_view_path["path"].format(**template)}'

    def build_url_parquet_staging(self, staging_name, connector_id):
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            staging_type=staging_name,
            connector_id=connector_id,
            tenant_id=tenant_id,
            bucket=self.carolina.cds_app_storage_path["bucket"]
        )
        return f's3://{self.carolina.cds_app_storage_path["bucket"]}/{self.carolina.cds_staging_path["path"].format(**template)}'

    def build_url_parquet_staging_master(self, staging_name, connector_id):
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            staging_type=staging_name,
            connector_id=connector_id,
            tenant_id=tenant_id,
            bucket=self.carolina.cds_app_storage_path["bucket"]
        )
        return f's3://{self.carolina.cds_app_storage_path["bucket"]}/{self.carolina.cds_staging_master_path["path"].format(**template)}'

    def build_url_parquet_staging_rejected(self, staging_name, connector_id):
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            staging_type=staging_name,
            connector_id=connector_id,
            tenant_id=tenant_id,
            bucket=self.carolina.cds_app_storage_path["bucket"]
        )
        return f's3://{self.carolina.cds_app_storage_path["bucket"]}/{self.carolina.cds_staging_rejected_path["path"].format(**template)}'

    def get_dask_options(self):
        return {"key": self.carolina.token['aiAccessKeyId'],
                "secret": self.carolina.token['aiSecretKey'],
                "token": self.carolina.token['aiAccessToken']}

    def get_golden_file_paths(self, dm_name):
        self._init_if_needed()
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            dm_name=dm_name,
            tenant_id=tenant_id
        )
        parq = list(
            self.bucket.objects.filter(Prefix=self.carolina.cds_golden_path['path'].format(**template))
        )
        return [{'storage_space': 'golden', 'name': i.key} for i in parq if i.key.endswith('.parquet')]

    def get_view_file_paths(self, view_name):
        self._init_if_needed()
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            relationship_view_name=view_name,
            tenant_id=tenant_id
        )
        parq = list(
            self.bucket.objects.filter(Prefix=self.carolina.cds_view_path['path'].format(**template))
        )
        return [{'storage_space': 'view', 'name': i.key} for i in parq if i.key.endswith('.parquet')]

    def get_staging_file_paths(self, staging_name, connector_id):
        self._init_if_needed()
        tenant_id = self.carol.tenant['mdmId']
        template = dict(
            staging_type=staging_name,
            connector_id=connector_id,
            tenant_id=tenant_id
        )
        parq_stag = list(
            self.bucket.objects.filter(Prefix=self.carolina.cds_staging_path['path'].format(**template))
        )
        parq_master = list(
            self.bucket.objects.filter(Prefix=self.carolina.cds_staging_master_path['path'].format(**template))
        )
        parq_stag_rejected = list(
            self.bucket.objects.filter(Prefix=self.carolina.cds_staging_rejected_path['path'].format(**template))
        )

        bs = [{'storage_space': 'staging', 'name': i.key} for i in parq_stag if i.key.endswith('.parquet')]
        bm = [{'storage_space': 'staging_master', 'name': i.key} for i in parq_master if i.key.endswith('.parquet')]
        br = [{'storage_space': 'staging_rejected', 'name': i.key} for i in parq_stag_rejected if i.key.endswith('.parquet')]

        return bs + bm + br


    def files_storage_list(self, app_name=None, all_apps=False,  print_paths=False):
        """

        It will return all files in Carol data Storage (CDS).


        :param app_name: `str`, default `None`
            app_name to filter output. If 'None' it will get value used to initialize `Carol()`
        :param all_apps: `bool`, default `False`
            Get all files in CDS.
        :param print_paths: `bool`, default `False`
            Print the tree structure of the files in CDS
        :return: list of files paths.
        """

        split = f"storage/{self.carol.tenant['mdmId']}/"
        if all_apps:
            prefix = f"storage/{self.carol.tenant['mdmId']}/"

        elif app_name is None:
            app_name = self.carol.app_name
            prefix = f"storage/{self.carol.tenant['mdmId']}/{app_name}/files/"

        else:
            prefix = f"storage/{self.carol.tenant['mdmId']}/{app_name}/files/"

        self._init_if_needed()

        files = list(self.bucket.objects.filter(Prefix=prefix))
        files = [i.key.split(split)[1] for i in files]

        if print_paths:
            main_dict = defaultdict(dict, ((_FILE_MARKER, []),))
            for line in files:
                _attach_path(line, main_dict)
            prettify_path(main_dict)
        return files



class Entity:
    
    def __init__(self, name, values, canonical_value=None):
        self.name = name
        self.values = values
        self.canonical_value = canonical_value

        self._json_data = {}
        self._update_json_data()

    @classmethod
    def from_json(cls, json_data):
        return cls(name=json_data['nlpName'],
                   values=json_data.get('nlpValues', []),
                   canonical_value=json_data.get('nlpCanonicalValue'))

    def add_values(self, values):
        if isinstance(values, list):
            self.values.extend(values)
        else:
            self.values.append(values)
        self._json_data['nlpValues'] = self.values
        
    def _update_json_data(self):
        self._json_data['nlpName'] = self.name
        self._json_data['nlpValues'] = self.values
        self._json_data['nlpCanonicalValue'] = self.canonical_value
        self._json_data.pop('mdmCreated',None)
        self._json_data.pop('mdmId',None)
        self._json_data.pop('mdmLastUpdated',None)
        self._json_data.pop('mdmTenantId',None)




"""
PyCarol Pipeline
=================================

This module defines a framework for building and managing data pipelines. It was inspired in Luigi's architecture and
has it as a backend for the implementation of the pipeline execution.

"""
from .task import (
    Task,
    WrapperTask,
    inherit_list,
    inherit_dict
)

from .targets import (
    CDSTarget,
    PickleTarget,
    KerasTarget,
    DummyTarget,
    JsonTarget,
    FeatherTarget,
    PytorchTarget,
    ParquetTarget,
    FileTarget,
    LocalTarget
)
from .targets.deprecated_targets import (
    PicklePyCarolTarget,
    PytorchPyCarolTarget,
    KerasPyCarolTarget,
    PickleLocalTarget,
    KerasLocalTarget,
    PytorchLocalTarget,
    JsonLocalTarget,
    FeatherLocalTarget,
    ParquetPyCarolTarget
)

from .tools import (
    Pipe,
)


import os


class Tasks:
    def __init__(self, carol):
        self.carol = carol

        self.task_id = None
        self.mdm_data = None
        self.mdm_user_id = None
        self.mdm_connector_id = None
        self.mdm_task_ready = None
        self.mdm_task_processing = None
        self.mdm_task_status = None
        self.mdm_task_owner = None
        self.mdm_task_progress = None
        self.mdm_process_after = None
        self.mdm_distribution_value = None
        self.mdm_task_priority = None
        self.mdm_number_of_steps = None
        self.mdm_number_of_steps_executed = None
        self.mdm_entity_type = None
        self.mdm_created = None
        self.mdm_last_updated = None
        self.mdm_tenant_id = None

    def _set_task_by_json(self, json_task):
        self.task_id = json_task['mdmId']
        self.mdm_user_id = json_task['mdmUserId']
        self.mdm_connector_id = json_task['mdmConnectorId']
        self.mdm_task_ready = json_task['mdmTaskReady']
        self.mdm_task_processing = json_task['mdmTaskProcessing']
        self.mdm_task_status = json_task['mdmTaskStatus']
        self.mdm_task_owner = json_task['mdmTaskOwner']
        self.mdm_task_progress = json_task['mdmTaskProgress']
        self.mdm_process_after = json_task['mdmProcessAfter']
        self.mdm_distribution_value = json_task['mdmDistributionValue']
        self.mdm_task_priority = json_task.get('mdmTaskPreference')
        self.mdm_number_of_steps = json_task['mdmNumberOfSteps']
        self.mdm_number_of_steps_executed = json_task['mdmNumberOfStepsExecuted']
        self.mdm_entity_type = json_task['mdmEntityType']
        self.mdm_created = json_task['mdmCreated']
        self.mdm_last_updated = json_task['mdmLastUpdated']
        self.mdm_tenant_id = json_task['mdmTenantId']

        if 'mdmData' in json_task:
            self.mdm_data = json_task['mdmData']

    def create(self, task_type, task_group, data={}):
        """
        Create a new task
        :param task_type: type of task
        :param task_group: commonly is used tenandId
        :param data: data used in the task
        :return: Task
        """
        dataJson = {
            "mdmTaskType": task_type,
            "mdmTaskGroup": task_group,
            "mdmData": data,
        }

        json_task = self.carol.call_api('v1/tasks/new', data=dataJson)
        self._set_task_by_json(json_task)
        return self

    def current_task(self):
        task_id = os.environ['LONGTASKID']
        if task_id is None:
            print("Can only get current_task if being called by Carol as a batch app")
        self.get_task(task_id)

    def get_current_task_id(self):
        task_id = os.environ['LONGTASKID']
        return task_id

    def set_as_current_task(self):
        os.environ['LONGTASKID'] = self.task_id

    def get_task(self, task_id=None):
        """
        Get Task
        :param task_id: task id
        :return: Task
        """

        if task_id is None:
            task_id = self.task_id

        json_task = self.carol.call_api('v1/tasks/{}'.format(task_id))
        self._set_task_by_json(json_task)
        return self

    def trace(self, log_message, task_id=None):
        self.add_log(log_message, "TRACE", task_id)

    def debug(self, log_message, task_id=None):
        self.add_log(log_message, "DEBUG", task_id)

    def info(self, log_message, task_id=None):
        self.add_log(log_message, "INFO", task_id)

    def warn(self, log_message, task_id=None):
        self.add_log(log_message, "WARN", task_id)

    def error(self, log_message, task_id=None):
        self.add_log(log_message, "ERROR", task_id)

    def add_log(self, log_message, log_level="INFO", task_id=None):
        """
        Add a log
        :param log_message: commonly used tenandId
        :param log_level: options: ERROR, WARN, INFO, DEBUG, TRACE
        :param task_id: it's not necessary if self.mdm_id is defined or if we are running from Carol as a batch app
        :return: boolean
        """

        if task_id is None:
            task_id = self.task_id
        if task_id is None:
            task_id = self.get_current_task_id()

        log = [{
            "mdmTaskId": task_id,
            "mdmLogMessage": log_message,
            "mdmLogLevel": log_level.upper()
        }]
        return self.add_logs(log)

    def add_logs(self, logs, task_id=None):
        """
        Add more than one log

        Args:
        logs: `list`
            list of logs objects [{"task_id":"", "log_message": "", "log_level": ""}]
        task_id `str` default `None`
                The task ID. it's not necessary if self.task_id is defined

        :return: Task
        """

        if task_id is None:
            task_id = self.task_id

        resp = self.carol.call_api('v1/tasks/{}/logs'.format(task_id), data=logs)
        if resp['success']:
            return True
        else:
            return False

    def get_logs(self, task_id=None):
        """
        Get all logs

        Args:
           task_id `str` default `None`
                The task ID. it's not necessary if self.task_id is defined

        :return:
            list of logs
        """

        if task_id is None:
            task_id = self.task_id

        resp = self.carol.call_api('v1/tasks/{}/logs'.format(task_id))
        return resp

    def set_progress(self, progress, progress_data=None, task_id=None):
        """
        Set Task Progress

        Args:
            progress: `int`
                Number relative to progress
            progress_data: 'dict` default `None`
                Json payload to be sent to Carol
            task_id `str` default `None`
                The task ID. it's not necessary if self.task_id is defined

        :return:
            Task response.
        """

        if progress_data is None:
            progress_data = {}
        else:
            assert isinstance(progress_data, dict)

        if task_id is None:
            task_id = self.task_id

        resp = self.carol.call_api('v1/tasks/{}/progress/{}'.format(task_id, progress), data=progress_data)
        return resp

    def cancel(self, task_id=None, force=False):
        """
        Cancel the task

        Args:
            task_id: `str` default `None`
                The task ID. it's not necessary if self.task_id is defined
            force: `boll` default `False`
                Force cancel

        :return:
            boolean
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set"

        querystring = {"force": force}

        resp = self.carol.call_api('v1/tasks/{}/cancel'.format(task_id), method="POST",params=querystring )
        if resp['success']:
            return True
        else:
            return False

    def fail(self, task_id=None, message=''):
        """
        Fail the task

        Args:
            task_id: `str` default `None`
                The task Id. it's not necessary if self.task_id is defined
            :param message: `str` default ``
                message to log

        :return:
            boolean
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set"

        querystring = {"message": message}

        resp = self.carol.call_api('v1/tasks/{}/fail'.format(task_id), method="POST",params=querystring )
        if resp['success']:
            return True
        else:
            return False



import warnings
import pandas as pd
import json
from ..utils.importers import _import_dask, _import_pandas
from ..storage import Storage
from ..query import Query
from ..filter import TYPE_FILTER, Filter
import itertools
from ..utils.miscellaneous import drop_duplicated_parquet

class DataModelView:

    def __init__(self, carol):
        self.carol = carol

    def get_by_name(self, view_name):
        return self._get(view_name, by='name')

    def get_by_id(self, view_id):
        return self._get(view_id, by='id')


    def _get(self, id,  by='name'):

        if by == 'name':
            url = f"v1/relationshipView/name/{id}"
        elif by == 'id':
            url = f"/api/v1/relationshipView/{id}"
        else:
            raise print('Type incorrect, it should be "id" or "name"')

        return self.carol.call_api(url, method='GET')

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                 "sortBy": self.sort_by}

    def reprocess(self, view_name=None, view_id=None, cds=False):
        """
        Reprocess the view records.

        Args:
            view_name: `str`, default `None`
                View name
            view_id: `str`, default `None`
                View ID
            cds: `bool`, default `False`
                Save view on CDS.

        Returns: `dict`
            Task response.

        """

        if view_name is not None:
            view_id = self.get_by_name(view_name=view_name)['mdmId']
        else:
            assert view_id is not None, "'view_id' or 'view_name' must be set"

        url_filter = "v1/goldenRecordView/reprocess"

        query_params = {"relationshipViewId" : view_id,
                        "cds" : cds}
        return self.carol.call_api(url_filter, params=query_params,
                                   method='POST')

    def get_all(self, offset=0, page_size=-1, sort_order='ASC',
                sort_by=None, print_status=False,
                save_file=None):


        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self._build_query_params()

        self.template_dict = {}
        self.template_data = []
        count = self.offset

        set_param = True
        self.total_hits = float("inf")
        if save_file:
            assert isinstance(save_file, str)
            file = open(save_file, 'w', encoding='utf8')
        while count < self.total_hits:
            url_filter = "v1/relationshipView"
            query = self.carol.call_api(url_filter, params=self.query_params, method='GET')

            if query['count'] == 0:
                print('There are no more results.')
                print('Expecting {}, reponse = {}'.format(self.total_hits, count))
                break
            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.template_data.extend(query)

            self.template_dict.update({i['mdmName']: {'mdmId': i['mdmId'],
                                                      'mdmRunningState': i['mdmRunningState'],
                                                      'mdmEntityType': i['mdmEntityType']}
                                       for i in query})

            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()
        return self


    def export(self, view_name=None, view_id=None, sync_view=True, full_export=False,
               delete_previous=False, export_format='PARQUET'):
        """

        Export datamodel to s3

        This method will trigger or pause the export of the data in the datamodel to
        s3

        :param view_name: `str`, default `None`
            View Name
        :param view_id: `str`, default `None`
            View id
        :param sync_view: `bool`, default `True`
            Sync the view
        :param full_export: `bool`, default `True`
            Do a resync of the view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :param export_format: `str`, default `PARQUET`
            Format of the data to be exported. Possible values are:
                "PARQUET", "CSV", "JSON", "EXCEL"
        :return: None
        """

        if sync_view:
            status = 'RUNNING'
        else:
            status = 'PAUSED'

        if view_id:
            view_name = self.get_by_id(view_id)['mdmName']
        else:
            assert view_name

        query_params = {"status": status, "fullExport": full_export,
                        'viewName':view_name, 'format':export_format,
                        "deletePrevious": delete_previous}

        url = f'v1/goldenRecordView/exporter'
        return self.carol.call_api(url, method='POST', params=query_params)


    def fetch_parquet(self, view_name, merge_records=True, backend='pandas', return_dask_graph=False,
                      columns=None, return_metadata=False, callback=None, max_hits=None,
                      cds=False, max_workers=None,):

        """

        Args:
            view_name: `str`
                View name to be imported
            merge_records: `bool`, default `True`
                This will keep only the most recent record exported. Sometimes there are updates and/or deletions and
                one should keep only the last records.
            backend: ['dask','pandas'], default `dask`
                if to use either dask or pandas to fetch the data
            return_dask_graph: `bool`, default `false`
                If to return the dask graph or the dataframe.
            columns: `list`, default `None`
                List of columns to fetch.
            return_metadata: `bool`, default `False`
                To return or not the fields ['mdmId', 'mdmCounterForEntity']
            callback: `callable`, default `None`
                Function to be called each downloaded file.
            max_hits: `int`, default `None`
                Number of records to get.
            cds: `bool`, default `False`
                    Get view data from CDS.
            max_workers: `int` default `None`
                Number of workers to use when downloading parquet files with pandas back-end.
        :return:
        """

        if callback:
            assert callable(callback), \
                f'"{callback}" is a {type(callback)} and is not callable.'

        if isinstance(columns, str):
            columns = [columns]

        assert backend == 'dask' or backend == 'pandas'

        if return_dask_graph:
            assert backend == 'dask'


        dms = self._get_view_export_stats()

        if not cds:
            if not dms.get(view_name):
                raise Exception(
                    f'"{view_name}" is not set to export data, \n'
                    f'use `dm = DataModelView(login).export(view_name="{view_name}", sync_dm=True) to activate')
            import_type = 'view'
        else:
            import_type = 'view_cds'
        if columns:
            columns.extend(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated'])




        storage = Storage(self.carol)
        token_carolina = storage.backend.carolina.token
        storage_space = storage.backend.carolina.get_bucket_name(import_type)

        if backend == 'dask':
            d = _import_dask(storage=storage, view_name=view_name, import_type=import_type,
                             merge_records=merge_records, return_dask_graph=return_dask_graph,
                             columns=columns, )

        elif backend == 'pandas':

            d = _import_pandas(storage=storage, view_name=view_name, golden=True, columns=columns, callback=callback,
                               max_hits=max_hits, import_type=import_type, max_workers=max_workers,
                               token_carolina=token_carolina,
                               )
            if d is None:
                warnings.warn("No data to fetch!", UserWarning)
                return d

        else:
            raise ValueError(f'backend should be either "dask" or "pandas" you entered {backend}')

        if merge_records:
            if not return_dask_graph:
                d = drop_duplicated_parquet(d)
            else:
                d = d.set_index('mdmCounterForEntity', sorted=True) \
                    .drop_duplicates(subset='mdmId', keep='last') \
                    .reset_index(drop=True)

        if not return_metadata:
            to_drop = set(['mdmId', 'mdmCounterForEntity', 'mdmLastUpdated']).intersection(set(d.columns))
            d = d.drop(labels=to_drop, axis=1)

        return d


    def export_all(self, sync_view=True, full_export=False, delete_previous=False):
        """

        Export all datamodel to s3

        This method will trigger or pause the export of the data in the data model view to CDS

        :param sync_view: `bool`, default `True`
            Sync the data model view
        :param full_export: `bool`, default `True`
            Do a resync of the data model view
        :param delete_previous: `bool`, default `False`
            Delete previous exported files.
        :return: None
        """
        self.get_all()

        for _name, i in self.template_dict.items():
            view_id = i['mdmId']
            self.export(view_id=view_id, sync_view=sync_view, full_export=full_export,
                        delete_previous=delete_previous)


    def _get_view_export_stats(self):
        """
        Get export status for views

        :return: `dict`
            dict with the information of which data model view is exporting its data.
        """

        json_q = Filter.Builder(key_prefix="") \
            .must(TYPE_FILTER(value="mdmGoldenRecordViewExport")).build().to_json()

        query = Query(self.carol, index_type='CONFIG', page_size=1000, only_hits=False)
        query.query(json_q, ).go()

        dm_results = query.results
        dm_results = [elem.get('hits', elem) for elem in dm_results
                      if elem.get('hits', None)]
        dm_results = list(itertools.chain(*dm_results))

        dm = self.get_all().template_data
        dm = {i['mdmId']: i['mdmName'] for i in dm}

        if dm_results is not None:
            return {dm.get(i['mdmRelationshipViewId'], i['mdmRelationshipViewId'] + '_NOT_FOUND'): i for i in dm_results}

        return dm_results


